import logging
from typing import Annotated, List, Optional

import strawberry
import strawberry_django
from asgiref.sync import sync_to_async
from django.db.models import Q

from openbook_auth.models import User as UserModel
from openbook_auth.schema.types import User
from openbook_common.schema.types import Paged
from openbook_insights.models import Insight as InsightModel
from openbook_posts.models import Post as PostModel
from openbook_posts.models import PostComment as PostCommentModel
from openbook_terms.schema.types import Topic

logger = logging.getLogger(__name__)


@strawberry_django.type(InsightModel)
class Insight:
    id: strawberry.auto
    title: strawberry.auto
    description: strawberry.auto
    creator: Optional[User] = strawberry_django.field(select_related=["creator__profile"])
    date_published: strawberry.auto
    header_image: Optional[str]
    header_image_blurhash: Optional[str]
    topic: Optional[Topic]

    @strawberry_django.field()
    def posts(self, info) -> List[Annotated["Post", strawberry.lazy("openbook_posts.schema.types")]]:
        current_user = info.context.request.user
        return self.posts.filter(PostModel.is_visible_query(current_user)).order_by("created")

    # this should be removed and be replaced by paged posts
    @strawberry_django.field()
    def number_of_posts(self, info) -> int:
        current_user = info.context.request.user
        return self.posts.filter(PostModel.is_visible_query(current_user)).count()

    @strawberry_django.field(
        description="A list of all users involved. Curator of the insight, Authors of all the posts and all commenters of those posts"
    )
    def involved_users(self, info, offset: int = 0, limit: int = 3) -> Paged[User]:
        current_user = info.context.request.user
        query = UserModel.objects.filter(
            Q(
                Q(curated_insights=self)
                | (Q(posts__insights=self) & PostModel.is_visible_query(current_user, "posts"))
                | (
                    Q(posts_comments__post__insights=self)
                    & PostCommentModel.is_visible_query(current_user, "posts_comments")
                )
            ),
            is_deleted=False,
        ).distinct()
        query = query.select_related("profile").cache(ops=["fetch", "count"], timeout=300)
        return Paged.of(query, offset, limit)
