import uuid
from enum import Enum
from requests import Session
from requests.adapters import HTTPAdapter, Retry

from typing import List
from typing import Optional

from openbook_auth.models import User
from openbook_communities.models import Community, CommunityMembershipRequest, CommunityInvite

from openbook_notifications.helpers import flatten_payload
from openbook_notifications.payloads import (
    SpaceMembershipRequestPayload,
    SpaceMembershipRequestApprovalUserPayload,
    SpaceMembershipRequestApprovalAdminPayload,
    SpaceMembershipRequestDeclinePayload,
    NotificationType,
    SpaceMembershipInvitePayload,
)
from openbook_notifications.toggles import notifications_toggle

from novu.api import EventApi


class Workflow(Enum):
    SPACE_MEMBERSHIP_REQUEST = "space-membership-request"
    SPACE_MEMBERSHIP_REQUEST_APPROVAL_USER = "space-membership-request-approval-user"
    SPACE_MEMBERSHIP_REQUEST_APPROVAL_ADMIN = "space-membership-request-approval-admin"
    SPACE_MEMBERSHIP_REQUEST_DECLINE_USER = "space-membership-request-decline-user"
    SPACE_MEMBERSHIP_REQUEST_DECLINE_ADMIN = "space-membership-request-decline-admin"
    SPACE_MEMBERSHIP_INVITE = "space-membership-invite"


# Use this classes static methods to trigger notifications via the Notifications system (novu)
# To avoid circular import issues (due to usage of domain classes here such as User) import deferred in method bodies
class Notifications:
    @staticmethod
    def send_space_membership_request_notification(
        membership_request: CommunityMembershipRequest,
        recipients: User | List[User],
        requester: User,
        space: Community,
        message: Optional[str] = "",
    ):
        notification = SpaceMembershipRequestNotification(
            recipients=[str(recipient.username) for recipient in recipients],
            payload=SpaceMembershipRequestPayload(space=space, user=requester, message=message),
            transaction_id=str(membership_request.id),
        )
        notification.send()

    @staticmethod
    def send_space_membership_request_approval_user_notification(
        requester: User,
        space: Community,
        approver: User,
    ):
        notification = SpaceMembershipRequestApprovalUserNotification(
            recipient=requester.username,
            payload=SpaceMembershipRequestApprovalUserPayload(user=requester, space=space, approver=approver),
        )
        notification.send()

    @staticmethod
    def send_space_membership_request_approval_admin_notification(
        recipients: User | List[User],
        requester: User,
        space: Community,
    ):
        notification = SpaceMembershipRequestApprovalAdminNotification(
            recipients=[str(recipient.username) for recipient in recipients],
            payload=SpaceMembershipRequestApprovalAdminPayload(user=requester, space=space),
        )
        notification.send()

    @staticmethod
    def send_space_membership_request_decline_user_notification(
        requester: User,
        space: Community,
        denier: User,
    ):
        notification = SpaceMembershipRequestDeclineUserNotification(
            recipient=requester.username,
            payload=SpaceMembershipRequestDeclinePayload(
                user=requester,
                space=space,
                denier=denier,
                notification_type=NotificationType.SPACE_MEMBERSHIP_REQUEST_DECLINE_USER.value,
            ),
        )
        notification.send()

    @staticmethod
    def send_space_membership_request_decline_admin_notification(
        recipients: User | List[User],
        requester: User,
        space: Community,
        denier: User,
    ):
        notification = SpaceMembershipRequestDeclineAdminNotification(
            recipients=[str(recipient.username) for recipient in recipients],
            payload=SpaceMembershipRequestDeclinePayload(
                user=requester,
                space=space,
                denier=denier,
                notification_type=NotificationType.SPACE_MEMBERSHIP_REQUEST_DECLINE_ADMIN.value,
            ),
        )
        notification.send()

    @staticmethod
    def send_space_membership_invite_notification(
        community_invite: CommunityInvite,
        recipient: User,
        inviter: User,
        space: Community,
    ):
        notification = SpaceMembershipInviteNotification(
            recipient=str(recipient.username),
            payload=SpaceMembershipInvitePayload(user=inviter, space=space),
            transaction_id=str(community_invite.id),
        )
        notification.send()


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class Notification:
    def __init__(self, workflow: str, recipients: str | List[str], payload: dict, transaction_id: str):
        session = Session()
        retries = Retry(
            total=5,  # Number of retries
            backoff_factor=1,  # Delay between retries (exponential backoff)
            status_forcelist=[502, 503, 504],  # Include common error codes
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)

        self.workflow = workflow
        self.recipients = recipients
        self.payload = payload
        self.session = session
        self.transaction_id = transaction_id

        self.overrides = {
            "type": "data",  # data notification
            "fcm": {"type": "data", "data": flatten_payload(payload)},
            "apns": {"payload": payload},
        }

    @notifications_toggle
    def send(self):
        return EventApi(session=self.session).trigger(
            name=self.workflow,
            recipients=self.recipients,
            payload=self.payload,
            overrides=self.overrides,
            transaction_id=self.transaction_id,
        )


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class SpaceMembershipRequestNotification(Notification):
    def __init__(self, recipients: str | List[str], payload: SpaceMembershipRequestPayload, transaction_id: str):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_REQUEST.value,
            recipients=recipients,
            payload=payload.to_dict(),
            transaction_id=transaction_id,
        )


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class SpaceMembershipRequestApprovalUserNotification(Notification):
    def __init__(
        self, recipient: str, payload: SpaceMembershipRequestApprovalUserPayload, transaction_id: Optional[str] = None
    ):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_REQUEST_APPROVAL_USER.value,
            recipients=recipient,
            payload=payload.to_dict(),
            transaction_id=transaction_id if transaction_id else str(uuid.uuid4()),
        )


class SpaceMembershipRequestApprovalAdminNotification(Notification):
    def __init__(
        self,
        recipients: str | List[str],
        payload: SpaceMembershipRequestApprovalAdminPayload,
        transaction_id: Optional[str] = None,
    ):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_REQUEST_APPROVAL_ADMIN.value,
            recipients=recipients,
            payload=payload.to_dict(),
            transaction_id=transaction_id if transaction_id else str(uuid.uuid4()),
        )


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class SpaceMembershipRequestDeclineUserNotification(Notification):
    def __init__(
        self, recipient: str, payload: SpaceMembershipRequestDeclinePayload, transaction_id: str | None = None
    ):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_REQUEST_DECLINE_USER.value,
            recipients=recipient,
            payload=payload.to_dict(),
            transaction_id=transaction_id if transaction_id else str(uuid.uuid4()),
        )


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class SpaceMembershipRequestDeclineAdminNotification(Notification):
    def __init__(
        self,
        recipients: str | List[str],
        payload: SpaceMembershipRequestDeclinePayload,
        transaction_id: str | None = None,
    ):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_REQUEST_DECLINE_ADMIN.value,
            recipients=recipients,
            payload=payload.to_dict(),
            transaction_id=transaction_id if transaction_id else str(uuid.uuid4()),
        )


# INTERNAL USE ONLY, use Notifications instead to trigger notifications
class SpaceMembershipInviteNotification(Notification):
    def __init__(self, recipient: str, payload: SpaceMembershipInvitePayload, transaction_id: str):
        super().__init__(
            workflow=Workflow.SPACE_MEMBERSHIP_INVITE.value,
            recipients=recipient,
            payload=payload.to_dict(),
            transaction_id=transaction_id,
        )
